# -*- coding: utf-8 -*-
from collections.abc import Callable


class Infix:
    """
    Implements a custom Infix operator.

    Usage
    -----
        x |op| y    or    x <<op>> y

    Example
    -------
        >>> x = Infix(lambda x, y: x * y)
        >>> print(2 | x | 4)
        8

        >>> x = Infix(lambda x, y: x + y)
        >>> print(2 << x >> 4)
        6
    """

    def __init__(self, function: Callable):
        self.function = function

    def __ror__(self, other):
        return Infix(lambda x, self=self, other=other: self.function(other, x))

    def __or__(self, other):
        return self.function(other)

    def __rlshift__(self, other):
        return Infix(lambda x, self=self, other=other: self.function(other, x))

    def __rshift__(self, other):
        return self.function(other)

    def __call__(self, value1, value2):
        return self.function(value1, value2)


if __name__ == '__main__':

    x = Infix(lambda x, y: x * y)
    print(2 | x | 4)

    x = Infix(lambda x, y: x + y)
    print(2 << x >> 4)

    # join = Infix(lambda x,y: x+y)
    # print(2 |join| 4)

    # def curry(f,x):
    #     def curried_function(*args, **kw):
    #         return f(*((x,)+args),**kw)
    #     return curried_function
    # curry = Infix(curry)



# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod


__all__ = ["ABCMeta_Weak", "ABCMeta_Strong", "ABCMeta_Safe", "ABC_Strong",
           "ABC_Weak", "ABC_Safe"]


class ABCMeta_Weak(ABCMeta):
    """
    Standard python metaclass. It follows weak abstraction in the meaning, that
    it is enough to implement the abstarcts in the instance, they impose no
    restriction on the inherited class itself. Therefore, error only occurs at
    runtime.
    """

    @staticmethod
    def _get_cls_methods(namespace, nomagic=False):
        if not nomagic:
            return {name for name, value in namespace.items()
                    if callable(value)}
        else:
            def cond(n, v): return callable(v) and ('__' not in n)
            return {name for name, value in namespace.items()
                    if cond(name, value)}


class ABC_Weak(metaclass=ABCMeta_Weak):
    """
    Helper class that provides a standard way to create an ABC using
    inheritance.
    """
    __slots__ = ()


if __name__ == '__main__':

    """
    Explicit notation of an abstract method in a base class is
    not necessary. Error only occurs at runtime.
    """

    class ABC_Parent_Weak(ABC_Weak):

        @abstractmethod
        def abcParentWeak(self):
            pass

    class ABC_Child_Weak(ABC_Parent_Weak):

        @abstractmethod
        def abcChildWeak(self):
            pass

    class ABC_GrandChild_Weak(ABC_Child_Weak):

        @abstractmethod
        def abcGrandChildWeak(self):
            pass

    class GrandChild_Weak(ABC_GrandChild_Weak):
        """
        Comment out any of these implementations and witness how
        error arises on the line of object creation, but not during
        compilation.
        """

        def abcParentWeak(self):
            pass

        def abcChildWeak(self):
            pass

        def abcGrandChildWeak(self):
            pass

    grandchild_Weak = GrandChild_Weak()


class ABCMeta_Strong(ABCMeta_Weak):
    """
    Strong Python metaclass. It follows strong abstraction in the meaning, that
    an abstract function of any of the base classes must be implemented or
    delayed with the use of another @abstractmethod decorator.
    Error occurs at compilation time.
    """

    def __init__(self, name, bases, namespace, *args, **kwargs):
        super().__init__(name, bases, namespace, *args, **kwargs)

    def __new__(metaclass, name, bases, namespace, *args, **kwargs):
        cls = super().__new__(metaclass, name, bases, namespace, *args,
                              **kwargs)
        cls_methods = metaclass._get_cls_methods(namespace)
        for base in bases:
            base_abstracts = set()
            for name in getattr(base, "__abstractmethods__", set()):
                value = getattr(cls, name, None)
                if getattr(value, "__isabstractmethod__", False):
                    base_abstracts.add(name)
            for abstract in base_abstracts:
                if abstract not in cls_methods:
                    err_str = """Can't create abstract class {classname}!
                    {classname} must implement abstract method {method} of
                    class {base_class}!""".format(
                        classname=name,
                        method=abstract,
                        base_class=base.__name__)
                    raise TypeError(err_str)
        return cls


class ABC_Strong(metaclass=ABCMeta_Strong):
    """Helper class that provides a standard way to create an ABC using
    inheritance.
    """
    __slots__ = ()


class ABCMeta_Safe(ABCMeta_Weak):
    """
    Python metaclass for safe inheritance. Throws a TypeError
    if a method tries to shadow a definition in any of the base
    classes.
    """

    def __init__(self, name, bases, namespace, *args, **kwargs):
        super().__init__(name, bases, namespace, *args, **kwargs)

    def __new__(metaclass, name, bases, namespace, *args, **kwargs):
        cls = super().__new__(metaclass, name, bases, namespace, *args,
                              **kwargs)
        cls_methods = metaclass._get_cls_methods(namespace, nomagic=True)
        for base in bases:
            for method in cls_methods:
                if hasattr(base, method):
                    err_str = """Can't create abstract class {classname}!
                        Method {method} is already implemented in class
                        {base_class}!""".format(
                        classname=name,
                        method=method,
                        base_class=base.__name__)
                    raise TypeError(err_str)
        return cls


class ABC_Safe(metaclass=ABCMeta_Safe):
    """
    Helper class that provides a standard way to create an ABC using
    inheritance.
    """
    __slots__ = ()


if __name__ == '__main__':

    """
    The code below throws a TypeError if an abstract method is not
    included in the definition of child class.
    Error occurs at compilation time.
    """

    class ABC_Parent_Strong(ABC_Strong):

        @abstractmethod
        def abcParentStrong(self):
            pass

    class ABC_Child_Strong(ABC_Parent_Strong):

        @abstractmethod
        def abcParentStrong(self):
            pass

        @abstractmethod
        def abcChildStrong(self):
            pass

    class ABC_GrandChild_Strong(ABC_Child_Strong):

        @abstractmethod
        def abcParentStrong(self):
            pass

        @abstractmethod
        def abcChildStrong(self):
            pass

        @abstractmethod
        def abcGrandChildStrong(self):
            pass

    class GrandChild_Strong(ABC_GrandChild_Strong):

        def abcParentStrong(self):
            pass

        def abcChildStrong(self):
            pass

        def abcGrandChildStrong(self):
            pass

    grandchild_Strong = GrandChild_Strong()

    class ABC_Parent_Safe(ABC_Safe):

        def funcParentSafe(self):
            pass

    class ABC_Child_Safe(ABC_Parent_Safe):

        def funcChildSafe(self):
            pass

    class ABC_GrandChild_Safe(ABC_Child_Safe):

        def _funcParentSafe(self):
            pass



# -*- coding: utf-8 -*-
from inspect import signature, Parameter
from typing import Any, Hashable
from typing import Generic as _GenericAlias


__all__ = ["Signature", "ishashable"]


def ishashable(obj):
    return isinstance(obj, Hashable)


class Signature(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        if 'abstract' in args:
            setattr(self, 'isabstract', True)
        else:
            setattr(self, 'isabstract', False)

    @classmethod
    def from_func(cls, funcobj, *axioms, **kwargs):

        if getattr(funcobj, 'isabstractoperation', False):
            sig = Signature('abstract', **kwargs)
        else:
            sig = Signature(**kwargs)

        sig['name'] = funcobj.__name__
        if len(axioms) > 0:
            sig['axioms'] = frozenset(axioms)
        else:
            sig['axioms'] = frozenset()

        params = signature(funcobj).parameters
        if 'args' in params or 'kwargs' in params:
            raise TypeError('Operation can only have a finite number '
                            'of arguments!')
        if 'self' in params or 'cls' in params:
            sig['arity'] = len(params)-1
        else:
            sig['arity'] = len(params)

        if sig['arity'] == 0:
            sig['dtype'] = [Any]
        else:
            sig['dtype'] = []
            for pname, param in params.items():
                if pname not in ['self', 'cls']:
                    if param.annotation is not Parameter.empty:
                        sig['dtype'].append(param.annotation)
                    else:
                        sig['dtype'].append(Any)

        annotations = funcobj.__annotations__
        if 'return' in annotations:
            sig['rtype'] = annotations['return']
        else:
            sig['rtype'] = Any

        return sig

    @classmethod
    def from_property(cls, funcobj, **kwargs):

        if funcobj.isabstractattribute:
            sig = Signature('abstract', **kwargs)
        else:
            sig = Signature(**kwargs)

        sig['name'] = funcobj.__name__

        annotations = funcobj.__annotations__
        if 'return' in annotations:
            sig['rtype'] = annotations['return']
        else:
            sig['rtype'] = Any

        return sig

    def compatible_func(self, other : 'Signature') -> bool:
        """
        Returns True if two instances are compatible in terms of
        domain type and result type.
        """
        if not isinstance(other, Signature) :
            return False

        #check domain types
        for type1, type2 in zip(self['dtype'], other['dtype']):
            if type1 == Any :
                continue
            elif isinstance(type1, _GenericAlias):
                if not isinstance(type2, _GenericAlias):
                    typeargs = type1.__dict__['__args__']
                    if type2 not in typeargs:
                        return False
                else:
                    if type1 != type2 :
                        return False
            else:
                if type1 != type2 :
                    return False

        #check result type
        if isinstance(self['rtype'], _GenericAlias):
            if not isinstance(other['rtype'], _GenericAlias):
                typeargs = self['rtype'].__dict__['__args__']
                if not other['rtype'] in typeargs :
                    return False
            else:
                if self['rtype'] != other['rtype']  :
                    return False
        elif self['rtype'] != Any:
            if self['rtype'] != other['rtype']  :
                return False

        return True

    def accepts_func(self, other : 'Signature') -> bool:
        """
        Returns True if other (implemented operation's signature) is
        compatible to self (abstract operation's signature).
        Every value to every key of self must be present in other.
        For example, here is where axioms are forced to match
        between two signatures.
        """
        if any([not self.isabstract, other.isabstract]):
            return False
        if not self.compatible_op(other) :
            return False
        if not self['name'] == other['name']:
            return False

        for key,value in self.items():
            if key not in ['rtype', 'dtype']:
                if key in other:
                    if isinstance(value, frozenset) and \
                            isinstance(other[key], frozenset):
                        if not value.issubset(other[key]):
                            return False
                    else:
                        if not value == other[key]:
                            return False
                else:
                    return False
        return True

    def update_func(self, other : 'Signature'):
        """
        Updates self with the content of other. Both must be abstracts.
        """
        assert isinstance(other, Signature)
        assert self.isabstract and other.isabstract
        assert self.compatible_op(other)
        for key,value in other.items():
            if key not in ['rtype', 'dtype']:
                if isinstance(value, frozenset):
                    if key in self and isinstance(self[key], frozenset):
                        s = set(self[key])
                        s.update(set(value))
                        self[key] = frozenset(s)
                    else:
                        self[key] = value
                else:
                    self[key] = value
        return

    def accepts_prop(self, value : Any = None) -> bool:
        """
        Returns True if other (implemented attribute's signature) is
        compatible to self (abstract attribute's signature).
        Name and result type must match.
        """
        if not self.isabstract:
            return False
        if value is None:
            return False

        #check result type
        if isinstance(self['rtype'], _GenericAlias):
            if not isinstance(type(value), _GenericAlias):
                typeargs = self['rtype'].__dict__['__args__']
                if not type(value) in typeargs:
                    return False
            else:
                if self['rtype'] != type(value):
                    return False
        elif self['rtype'] != Any:
            if self['rtype'] != type(value):
                return False
        return True

    def accepts_params(self, *args):
        raise NotImplementedError



# -*- coding: utf-8 -*-
from collections import OrderedDict
from copy import copy


class NestedOrderedDict(OrderedDict):
    """
    A nested dictionary hierarchy. In contrast with the
    `Library` class, this does not implement a default
    behaviour for missing keys.
    """

    separator = '::'

    def __init__(self, *args, parent=None, separator=None, key=None,
                 name=None, dtype=None, ftype=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent = None
        if parent is not None:
            assert isinstance(parent, dict)
            self.parent = parent
            self.separator = self.parent.separator
        if separator is not None:
            self.separator = separator
        self._key = key
        self._name = name
        self.dtype = dtype if dtype is not None else type(self)
        self.ftype = ftype if ftype is not None else type(self)
        self._key_to_address_mapper = None

    def parser(self, address=None, *args, **kwargs):
        """
        Iterates through the values of a dictionary.

        Usage
        -----
            (1) If the keyword 'inclusive' is provided as True, returns root
                as the first item.
            (2) If 'relative' is in args, addresses are measured relative to
                the calling Library.
        Notes
        -----
            (1) Returns all kinds of items.
        """
        if address is None:
            address = [] if 'relative' in args else self.address

        if kwargs.pop('inclusive', False):
            yield address, self.key, self

        for key, value in self.items():
            if isinstance(value, self.ftype):
                subaddress = copy(address)
                subaddress.append(key)
                yield subaddress, key, value
            elif isinstance(value, self.dtype):
                subaddress = copy(address)
                subaddress.append(key)
                for parsedata in value.parser(subaddress):
                    yield parsedata


def dictparser(d: dict = None, address=None, *args, **kwargs):
    """
    Iterates through all the values of a nested dictionary.

    Notes
    -----
        (1) Returns all kinds of items.
    """
    if address is None:
        address = []
    for key, value in d.items():
        subaddress = copy(address)
        subaddress.append(key)
        yield subaddress, key, value
        if isinstance(value, dict):
            for data in dictparser(value, subaddress):
                yield data


if __name__ == '__main__':

    d = NestedOrderedDict({0: 1, 1: 2})
    e = NestedOrderedDict({2: 3, 3: 4})
    f = NestedOrderedDict({22: 13, 23: 54})
    g = NestedOrderedDict({24: 8, 73: 11})
    e['f'] = f
    e['g'] = g
    d['e'] = e

    print(list(d.values()))

    def foo(d):
        def yielder(x):
            yield x
        #isdict = lambda v : isinstance(v, dict)
        def isnotdict(v): return not isinstance(v, dict)
        vals = d.values()
        nodicts = list(filter(isnotdict, vals))
        while True:
            [yielder(v) for v in nodicts]
            #dicts = list(filter(isdict, vals))



# -*- coding: utf-8 -*-
from dewloosh.core.types.library import Library
from typing import Callable
import numpy as np

__version__ = "1.0.0"
__description__ = "Common developer utilities and base classes to support other dewloosh packages."


def squeeze(default=True):
    def decorator(fnc: Callable):
        def inner(*args, **kwargs):
            if kwargs.get('squeeze', default):
                res = fnc(*args, **kwargs)
                if isinstance(res, tuple):
                    return list(map(np.squeeze, res))
                return np.squeeze(res)
            else:
                return fnc(*args, **kwargs)
        return inner
    return decorator




# -*- coding: utf-8 -*-

__all__ = ['Wrapper', 'wrapper', 'customwrapper', 'wrap']

NoneType = type(None)


class Wrapper:
    """
    Wrapper base class that
        (a) wraps an existing object at object creation provided as a keyword
            argument with `Wrapper.wrapkey`
        (b) wraps an existing object at object creation if it is a positional
            argument and an instance of `Wrapper.wraptype`
        (b) wraps the object Wrapper.wraptype(*args, **kwargs) if 
            `Wrapper.wraptype` is not None 
    """
    wrapkey = 'wrap'
    wraptype = NoneType

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.wrap(kwargs.get(self.wrapkey, None))

        if self._wrapped is None and self.wraptype is not NoneType:
            for arg in args:
                if isinstance(arg, self.wraptype):
                    self._wrapped = arg
                    break

            if self._wrapped is None:
                try:
                    if self.wraptype is not NoneType:
                        self._wrapped = self.wraptype(*args, **kwargs)
                except Exception:
                    raise ValueError("Wrapped class '{}' cannot be "
                                     "initiated with these "
                                     "arguments".format(
                                         self.wraptype.__name__))
        else:
            if self.wraptype is not NoneType:
                assert isinstance(self._wrapped, self.wraptype), \
                    "Wrong type, unable to wrap object : {}". \
                    format(self._wrapped)

    def wrap(self, obj=None):
        if self.wraptype is not NoneType:
            if isinstance(obj, self.wraptype):
                self._wrapped = obj
        else:
            self._wrapped = obj

    def wraps(self):
        return self._wrapped is not None

    def wrapped_obj(self):
        return self._wrapped

    def __hasattr__(self, attr):
        return any([attr in self.__dict__, 
                    attr in self._wrapped.__dict__])

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)
        try:
            return getattr(self._wrapped, attr)
        except Exception:
            raise AttributeError("'{}' object has no attribute \
                called {}".format(self.__class__.__name__, attr))

    def __getitem__(self, index):
        try:
            return super().__getitem__(index)
        except Exception:
            try:
                return self._wrapped.__getitem__(index)
            except Exception:
                raise TypeError("'{}' object is not "
                                "subscriptable".format(
                                    self.__class__.__name__))

    def __setitem__(self, index, value):
        try:
            return super().__setitem__(index, value)
        except Exception:
            try:
                return self._wrapped.__setitem__(index,value)
            except Exception:
                raise TypeError("'{}' object does not support "
                                "item assignment".format(
                                    self.__class__.__name__))


def customwrapper(*args, wrapkey='wrap', wraptype=NoneType, **kwargs):
    """
    Returns a class decorator turning a class type into a wrapper type, that
        (a) wraps an existing object at object creation provided as a keyword
            argument with wrapkey
        (b) wraps an existing object at object creation if it is a positional
            argument and an instance of wraptype
        (b) wraps the object wraptype(*args, **kwargs)
    """
    class BaseWrapperType(Wrapper): ...
    BaseWrapperType.wrapkey = wrapkey
    BaseWrapperType.wraptype = wraptype    
    def wrapper(BaseType):   
        class WrapperType(BaseWrapperType, BaseType):
            basetype = BaseType    
        return WrapperType
    return wrapper


def wrapper(BaseType : type):
    """
    Simple class decorator that turns a type into a wrapper with default
    behaviour.
    """
    class WrapperType(Wrapper, BaseType):
        basetype = BaseType     
    return WrapperType


def wrap(obj : object) -> Wrapper:
    """
    Wraps an object and returns the wrapper.
    """
    return Wrapper(wrap=obj)


if __name__ == '__main__' :
    
    class Wrapped:

        def foo(self):
            print('foo in object to be wrapped')

        def boo(self):
            print('boo in object to be wrapped')

    @customwrapper(wrapkey='wrapkey', wraptype=Wrapped)
    class CustomWrapper:

        def boo(self):
            print('boo in custom wrapper')
            
    @wrapper
    class DefaultWrapper:

        def boo(self):
            print('boo in default wrapper')
            
    
    class DefaultWrapper2(Wrapper):

        def boo(self):
            print('boo in default wrapper')

    cw = CustomWrapper(wrapkey=Wrapped())
    cw.foo()
    cw.boo()
    
    dw = DefaultWrapper(wrap=Wrapped())
    dw.foo()
    dw.boo()
    
    dw2 = DefaultWrapper(wrap=Wrapped())
    dw2.foo()
    dw2.boo()
    
    w = wrap(Wrapped())
    w.foo()
    w.boo()



# -*- coding: utf-8 -*-
from typing import Union, Callable
from collections import OrderedDict
from copy import deepcopy, copy
from dewloosh.core.tools import issequence


__all__ = ["DefaultDict", "OrderedDefaultDict", "NestedDict",
           "NestedOrderedDict", "dictparser"]


class DefaultDict(dict):
    """
    Implements a defaultdict.
    Source: http://stackoverflow.com/a/6190500/562769
    """

    def __init__(self, default_factory: Callable = None, *args, **kwargs):
        if (default_factory is not None and
                not hasattr(default_factory, '__call__')):
            raise TypeError('first argument must be callable')
        super().__init__(*args, **kwargs)
        if default_factory is not None:
            self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = self.default_factory,
        return type(self), args, None, None, self.items()

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        return type(self)(self.default_factory,
                          deepcopy(self.items()))

    def __repr__(self):
        return 'defaultdict(%s, %s)' % (self.default_factory,
                                        dict.__repr__(self))

    def items(self, *args, deep=False, return_address=False, **kwargs):
        if deep:
            if return_address:
                return dictparser(self)
            else:
                return parseitems(self)
        else:
            for k, v in super().items():
                yield k, v

    def values(self, *args, deep=False, **kwargs):
        if deep:
            for _, v in parseitems(self):
                yield v
        else:
            for v in super().values():
                yield v

    def keys(self, *args, deep=False, **kwargs):
        if deep:
            for k, _ in parseitems(self):
                yield k
        else:
            for k in super().keys():
                yield k

    @classmethod
    def default_factory(cls):
        return cls()


class OrderedDefaultDict(OrderedDict):
    """
    Implements an ordered version of a defaultdict.
    Source: http://stackoverflow.com/a/6190500/562769
    """

    def __init__(self, default_factory: Callable = None, *args, **kwargs):
        if (default_factory is not None and
                not isinstance(default_factory, Callable)):
            raise TypeError('first argument must be callable')
        super().__init__(*args, **kwargs)
        if default_factory is not None:
            self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = self.default_factory,
        return type(self), args, None, None, self.items()

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        return type(self)(self.default_factory,
                          deepcopy(self.items()))

    def __repr__(self):
        return 'OrderedDefaultDict(%s, %s)' % (self.default_factory,
                                               dict.__repr__(self))

    def items(self, *args, deep=False, return_address=False, **kwargs):
        if deep:
            if return_address:
                return dictparser(self)
            else:
                return parseitems(self)
        else:
            for k, v in super().items():
                yield k, v

    def values(self, *args, deep=False, **kwargs):
        if deep:
            for _, v in parseitems(self):
                yield v
        else:
            for v in super().values():
                yield v

    def keys(self, *args, deep=False, **kwargs):
        if deep:
            for k, _ in parseitems(self):
                yield k
        else:
            for k in super().keys():
                yield k

    @classmethod
    def default_factory(cls):
        return cls()


class DictCollection(DefaultDict):

    def __init__(self, *args, **kwargs):
        super().__init__(None, *args, **kwargs)

    def __repr__(self):
        return 'DictCollection(%s)' % dict.__repr__(self)

    def __getitem__(self, key):
        try:
            if issequence(key):
                return parseaddress(self, key)
            else:
                return super().__getitem__(key)
        except ValueError:
            return self.__missing__(key)
        except KeyError:
            return self.__missing__(key)

    def __setitem__(self, key, value):
        try:
            if issequence(key):
                if not key[0] in self:
                    d = self.__missing__(key[0])
                else:
                    d = self[key[0]]
                if len(key) > 1:
                    d.__setitem__(key[1:], value)
                else:
                    self[key[0]] = value
            else:
                return super().__setitem__(key, value)
        except AttributeError:
            raise RuntimeError(
                "Target is of type '{}', which is not a container.".format(type(d)))
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if issequence(key):
            if key[0] not in self:
                self[key[0]] = value = self.default_factory()
            else:
                value = self[key[0]]
            if len(key) > 1:
                return value.__missing__(key[1:])
            else:
                return value
        else:
            self[key] = value = self.default_factory()
            return value


class OrderedDictCollection(OrderedDefaultDict):

    def __init__(self, *args, **kwargs):
        super().__init__(None, *args, **kwargs)

    def __repr__(self):
        return 'OrderedDictCollection(%s)' % dict.__repr__(self)

    def __getitem__(self, key):
        try:
            if issequence(key):
                return parseaddress(self, key)
            else:
                return super().__getitem__(key)
        except ValueError:
            return self.__missing__(key)
        except KeyError:
            return self.__missing__(key)

    def __setitem__(self, key, value):
        try:
            if issequence(key):
                if not key[0] in self:
                    d = self.__missing__(key[0])
                else:
                    d = self[key[0]]
                if len(key) > 1:
                    d.__setitem__(key[1:], value)
                else:
                    self[key[0]] = value
            else:
                return super().__setitem__(key, value)
        except AttributeError:
            raise RuntimeError(
                "Target is of type '{}', which is not a container.".format(type(d)))
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if issequence(key):
            if key[0] not in self:
                self[key[0]] = value = self.default_factory()
            else:
                value = self[key[0]]
            if len(key) > 1:
                return value.__missing__(key[1:])
            else:
                return value
        else:
            self[key] = value = self.default_factory()
            return value


def NestedDict(ordered: bool = False) -> Union[DefaultDict,
                                               OrderedDefaultDict]:
    """
    Returns a nested default dictionary.
    """
    if ordered:
        return OrderedDefaultDict(lambda *_: NestedDict(ordered))
    else:
        return DefaultDict(lambda *_: NestedDict(ordered))


def NestedOrderedDict() -> OrderedDefaultDict:
    """
    Returns a nested and ordered default dictionary.
    """
    return OrderedDefaultDict(lambda *_: NestedOrderedDict())


def parseaddress(d: dict, a: list):
    if not isinstance(d, dict):
        raise ValueError
    if not a[0] in d:
        raise KeyError(a[0])
    if len(a) > 1:
        return parseaddress(d[a[0]], a[1:])
    else:
        return d[a[0]]


def parseitems(d: dict = None, *args, dtype=dict, **kwargs):
    """
    A generator function that yields all the items of a nested dictionary as
    (key, value) pairs.

    Notes
    -----
        Does not return nested dictionaries themselves, only their content.
    """
    for key, value in d.items():
        if isinstance(value, dtype):
            for data in parseitems(value, dtype=dtype):
                yield data
        else:
            yield key, value


def dictparser(d: dict = None, address=[], *args, dtype=dict, **kwargs):
    """
    Iterates through all the values of a nested dictionary.

    Notes
    -----
        Returns all kinds of items, even nested discionaries themselves,
        along with their content.
    """
    for key, value in d.items():
        subaddress = copy(address)
        subaddress.append(key)
        if isinstance(value, dtype):
            for data in dictparser(value, subaddress, dtype=dtype):
                yield data
        else:
            yield subaddress, value
            

def parsedicts_addr(d: dict = None, address=[], *args, inclusive=True, 
                    dtype=dict, deep=True, **kwargs):
    if inclusive:
        if isinstance(d, dtype):
            yield address, d
    for key, value in d.items():
        if isinstance(value, dtype):
            addr = copy(address)
            addr.append(key)
            yield addr, value
            if deep:
                for subaddr, subval in \
                    parsedicts_addr(value, addr, 
                                    inclusive=False, dtype=dtype):
                        yield subaddr, subval


def parsedicts(d: dict = None, *args, inclusive=True, dtype=dict,
               deep=True, **kwargs):
    if inclusive:
        if isinstance(d, dtype):
            yield d
    for value in d.values():
        if isinstance(value, dtype):
            yield value
            if deep:
                for subvalue in \
                    parsedicts(value, inclusive=False, dtype=dtype):
                        yield subvalue


if __name__ == '__main__':

    factory = lambda *_: 'none'

    dd = DefaultDict(factory)
    print(dd['key'])

    ndd = NestedDict(ordered=False)
    ndd['a']['b']['c']['e'] = 1
    ndd['a']['b']['d'] = 2
    print(ndd['a']['b']['c'])
    print(ndd['a']['b']['d'])

    for k, v in ndd.items():
        print((k, v))

    for v in ndd.values():
        print(v)

    for k in ndd.keys():
        print(k)

    for k in ndd.keys(deep=True):
        print(k)

    for k, v in ndd.items(deep=True):
        print((k, v))

    for a, v in ndd.items(deep=True, return_address=True):
        print((a, v))

    def foo(*a):
        print(type(a))

    dc = OrderedDictCollection()
    dc['a']['b']['c']['e'] = 1
    dc['a']['b']['d'] = 2

    for v in dc.values():
        print(v)

    for k in dc.keys():
        print(k)

    for k in dc.keys(deep=True):
        print(k)
